from __future__ import annotations

from .ppca_rs import *

from dataclasses import dataclass
from typing import List, Optional

import numpy as np

@dataclass(frozen=True)
class TrainMetrics:
    llk_per_sample: float
    aic: float
    bic: float


@dataclass
class PPCATrainer:
    """A trainer for a PPCA Model over masked data."""

    # The list of masked samples against which the PPCA will be trained.
    dataset: Dataset

    def __init(self, state_size: int) -> PPCAModel:
        """Initializes a first guess for the model."""
        init = PPCAModel.init(state_size, self.dataset)
        return init

    def train(
        self,
        *,
        start: Optional[PPCAModel] = None,
        state_size: int,
        n_iters: int = 10,
        quiet: bool = False,
    ) -> PPCAModel:
        """Trains a PPCA model for a given state size for a given number of iterations."""
        model = start or self.__init(state_size)

        for idx in range(n_iters):
            if not quiet:
                llk = model.llk(self.dataset)
                metrics = TrainMetrics(
                    llk_per_sample=llk / len(self.dataset),
                    aic=2.0 * (model.n_parameters - llk) / len(self.dataset),
                    bic=(llk - model.n_parameters * np.log(len(self.dataset)))
                    / len(self.dataset),
                )
                print(f"Masked PPCA iteration {idx + 1}: aic={metrics.aic}")

            model = model.iterate(self.dataset)

        return model.to_canonical()



import numpy as np

from ppca_rs import PPCAModel

real_model = PPCAModel(
    transform=np.matrix([[1, 1, 0], [1, 0, 1]], dtype="float64").T,
    isotropic_noise=0.1,
    mean=np.array([0, 1, 0], dtype="float64"),
)
sample = real_model.sample(100, mask_prob=0.2)
model = PPCAModel.init(2, sample)

for it in range(100):
    print(f"At iteration {it + 1} PPCA llk is {model.llk(sample)}")
    model: PPCAModel = model.iterate(sample)

print(model)



import numpy as np
from ppca_rs import PPCAModel

print("Generating model")

transform = np.random.binomial(1.0, 0.1, size=(200, 16))
real_model = PPCAModel(
    transform=transform.astype("float64"),
    isotropic_noise=0.1,
    mean=np.zeros((200, 1), dtype="float64"),
)

print("Generating syntetic sample")
sample = real_model.sample(100_000, 0.2)

print("Initializing model")
model = PPCAModel.init(16, sample)

print("Starting iterations...")

for it in range(24):
    print(f"At iteration {it + 1} PPCA llk is {model.llk(sample) / len(sample)}")
    model = model.iterate(sample)
    print("Done creating iterated model")

print("Model trained")



from typing import List, Optional

import numpy as np

class Dataset:
    """
    Represents a dataset. This is a wrapper over a 2D array of dimensions
    (n_samples, n_features).

    ## Note

    All arrays involved have to be of data type `float32`.
    """

    def __init__(self, ndarray: np.ndarray) -> None: ...
    def numpy(self) -> np.ndarray:
        """Returns the underlying dataset as a 2D numpy array."""
    def __len__(self) -> int: ...
    def is_empty(self) -> bool: ...
    def empty_dimensions(self) -> List[int]:
        """
        Returns the dimensions which have only masked values for all samples in this
        dataset.
        """
    def output_size(self) -> Optional[int]:
        """
        Returns the size of each sample in this dataset, if the dataset is not empty.
        Else, returns `None`.
        """

class InferredMasked:
    """
    A class containing the result of the Bayesian inference step in `PPCAModel.infer`.
    """

    def states(self) -> np.ndarray:
        """The inferred mean value for each sample."""
    def covariances(self) -> List[np.ndarray]:
        """
        The covariance matrices for each sample.The covariances here can change from
        sample to sample, depending on the mask. If there is lots of masking in a sample,
        the covariance will be overall bigger.
        """

class PPCAModel:
    """
    A PPCA model: each sample for this model behaves according to the following
    statistical latent variable model.
    ```
    x ~ N(0; I(nxn))
    y = C * x + y0 + noise
    noise ~ N(0; sgima ^ 2 * I(mxm))
    ```
    Here, `x` is the latent state, y is the observed sample, that is an affine
    transformation of the hidden state contaminated by isotropic noise.

    ## Note

    All arrays involved have to be of data type `float32`.
    """

    def __init__(
        self, isotropic_noise: float, transform: np.ndarray, mean: np.ndarray
    ) -> None: ...

    transform: np.ndarray
    """The linear transformation from hidden state space to output space."""
    isotropic_noise: float
    """The standard deviation of the noise in the output space."""
    mean: np.ndarray
    """Then center of mass of the distribution in the output space."""
    output_size: int
    """The number of features for this model."""
    state_size: int
    """The number of hidden values for this model."""
    n_parameters: int
    """The total number of parameters involved in training (used for information criteria)."""
    transform: np.ndarray

    @classmethod
    def init(cls, n_states: int) -> PPCAModel:
        """Creates an uninformed model to seed the trainment."""
    def __repr__(self) -> str: ...
    def llk(self, dataset: Dataset) -> float:
        """
        Calculates the log-probability of a given masked dataset according to the current
        model.
        """
    def sample(self, dataset_size: int, mask_prob: float) -> Dataset:
        """
        Samples random outputs from the model and masks each entry according to a
        Bernoulli (coin-toss) distribution of proability `mask_prob` of erasing the
        generated value.
        """
    def infer(self, dataset: Dataset) -> InferredMasked:
        """Infers the hidden components for each sample in the dataset."""
    def filter_extrapolate(self, dataset: Dataset) -> Dataset:
        """
        Filters a dataset of samples and extrapolates the missing values with the most
        probable values.
        """
    def extrapolate(self, dataset: Dataset) -> Dataset:
        """Extrapolates the missing values with the most probable values."""
    def iterate(self, dataset: Dataset) -> PPCAModel:
        """
        Makes one iteration of the EM algorithm for the PPCA over an onbserved dataset,
        returning the improved model.
        """
    def to_canonical(self) -> PPCAModel:
        """
        Returns a canonical version of this model. This does not alter the log-probablility
        function nor the quality of the training. All it does is to transform the hidden
        variables.
        """



import pickle
import numpy as np

from ppca_rs import PPCAModel

model = PPCAModel(
    transform=np.matrix([[1, 1, 0], [1, 0, 1]], dtype="float64").T,
    isotropic_noise=0.1,
    mean=np.array([0, 1, 0], dtype="float64"),
)

ser = pickle.dumps(model)
de = pickle.loads(ser)

print(model)
print(de)
